-- **************************************************
-- Provide Moho with the name of this script object
-- **************************************************

ScriptName = "FA_export_csv"

-- **************************************************
-- General information about this script
-- **************************************************

FA_export_csv = {}

function FA_export_csv:Name()
    return "Export CSV"
end

function FA_export_csv:Version()
    return "1.1"
end

function FA_export_csv:Description()
    return "Export to TVPaint's CSV layered animation export format"
end

function FA_export_csv:Creator()
    return "Fazek"
end

function FA_export_csv:UILabel()
    return("Export CSV animation...")
end

-- **************************************************
-- The guts of this script
-- **************************************************

FA_export_csv.filePath = ""
FA_export_csv.dirSep = "\\"

function FA_export_csv:OnMouseDown(moho, mouseEvent)
    self:Run(moho)
end

function FA_export_csv:IsAnimCh(channel)
    if channel ~= nil then
        if channel:CountKeys() > 1 or channel:CountActions() > 0 then
            if channel.IsMuted ~= nil then
                if channel:IsMuted() then
                    return false
                end
            end
            return true
        end
    end
    return false
end

-- if a layer is animated in an incompatible way, its all frames will be rendered separately

function FA_export_csv:IsAnimated(moho, layer)
    --check for animated visual features
    if self:IsAnimCh(moho.document.fCameraPanTilt) or
       self:IsAnimCh(moho.document.fCameraRoll) or
       self:IsAnimCh(moho.document.fCameraTrack) or
       self:IsAnimCh(moho.document.fCameraZoom) then
           return true
    end
    if self:IsAnimCh(layer.fAlpha) or
       self:IsAnimCh(layer.fBlur) or
       self:IsAnimCh(layer.fFlipH) or
       self:IsAnimCh(layer.fFlipV) or
       self:IsAnimCh(layer.fFollowing) or
       self:IsAnimCh(layer.fLayerColor) or
       self:IsAnimCh(layer.fLayerColorOn) or
       self:IsAnimCh(layer.fLayerOutline) or
       self:IsAnimCh(layer.fLayerShading) or
       self:IsAnimCh(layer.fLayerShadow) or
       self:IsAnimCh(layer.fMotionBlur) or
       self:IsAnimCh(layer.fMotionBlurAlphaEnd) or
       self:IsAnimCh(layer.fMotionBlurAlphaStart) or
       self:IsAnimCh(layer.fMotionBlurFrames) or
       self:IsAnimCh(layer.fMotionBlurRadius) or
       self:IsAnimCh(layer.fMotionBlurSkip) or
       self:IsAnimCh(layer.fOutlineColor) or
       self:IsAnimCh(layer.fOutlineWidth) or
       self:IsAnimCh(layer.fPerspectiveBlur) or
       self:IsAnimCh(layer.fPerspectiveColor) or
       self:IsAnimCh(layer.fPerspectiveScale) or
       self:IsAnimCh(layer.fPerspectiveShadow) or
       self:IsAnimCh(layer.fPerspectiveShear) or
       self:IsAnimCh(layer.fPhysicsNudge) or
       self:IsAnimCh(layer.fRotationX) or
       self:IsAnimCh(layer.fRotationY) or
       self:IsAnimCh(layer.fRotationZ) or
       self:IsAnimCh(layer.fScale) or
       self:IsAnimCh(layer.fShadingAngle) or
       self:IsAnimCh(layer.fShadingBlur) or
       self:IsAnimCh(layer.fShadingColor) or
       self:IsAnimCh(layer.fShadingContraction) or
       self:IsAnimCh(layer.fShadingNoiseAmp) or
       self:IsAnimCh(layer.fShadingNoiseScale) or
       self:IsAnimCh(layer.fShadingOffset) or
       self:IsAnimCh(layer.fShadowAngle) or
       self:IsAnimCh(layer.fShadowBlur) or
       self:IsAnimCh(layer.fShadowColor) or
       self:IsAnimCh(layer.fShadowExpansion) or
       self:IsAnimCh(layer.fShadowNoiseAmp) or
       self:IsAnimCh(layer.fShadowNoiseScale) or
       self:IsAnimCh(layer.fShadowOffset) or
       self:IsAnimCh(layer.fShear) or
       self:IsAnimCh(layer.fTranslation) then
           return true
    end

    if layer.fAmbientOcclusionRadius ~= nil then
        if self:IsAnimCh(layer.fAmbientOcclusionRadius) or
           self:IsAnimCh(layer.fMotionBlurExtendedFrames) or
           self:IsAnimCh(layer.fMotionBlurFramePercentage) or
           self:IsAnimCh(layer.fMotionBlurSubframes) or
           self:IsAnimCh(layer.fNoise) or
           self:IsAnimCh(layer.fPixelation) or
           self:IsAnimCh(layer.fThreshold) then
               return true
        end
    end

    if layer:LayerType() == MOHO.LT_VECTOR then
        local vec = moho:LayerAsVector(layer)
        if vec.fAnimatedNoise then
            return true
        end
        local mesh = vec:Mesh()
        local cnt = mesh:CountPoints()
        while cnt > 0 do
            cnt = cnt - 1
            local point = mesh:Point(cnt)
            if self:IsAnimCh(point.fAnimPos) or
               self:IsAnimCh(point.fColor) or
               self:IsAnimCh(point.fColorStrength) or
               self:IsAnimCh(point.fWidth) then
                   return true
            end
        end
        cnt = mesh:CountCurves()
        while cnt > 0 do
            cnt = cnt - 1
            local curve = mesh:Curve(cnt)
            if self:IsAnimCh(curve.fEndPercent) or
               self:IsAnimCh(curve.fStartPercent) then
                   return true
            end
        end
        cnt = mesh:CountShapes()
        while cnt > 0 do
            cnt = cnt - 1
            local shape = mesh:Shape(cnt)
            if self:IsAnimCh(shape.f3DThickness) or
               self:IsAnimCh(shape.fEffectOffset) or
               self:IsAnimCh(shape.fEffectRotation) or
               self:IsAnimCh(shape.fEffectScale) then
                   return true
            end
            local style = shape.fMyStyle
            if style ~= nil then
                if self:IsAnimCh(style.fFillCol) or
                    self:IsAnimCh(style.fLineCol) then
                        return true
                end
            end
            style = shape.fInheritedStyle
            if style ~= nil then
                if self:IsAnimCh(style.fFillCol) or
                    self:IsAnimCh(style.fLineCol) then
                        return true
                end
            end
            style = shape.fInheritedStyle2
            if style ~= nil then
                if self:IsAnimCh(style.fFillCol) or
                    self:IsAnimCh(style.fLineCol) then
                        return true
                end
            end
        end
    end
    return false
end

function FA_export_csv:CheckSwitch(moho, layer)

    if self:IsAnimated(moho, layer) then
        return 0
    end
    -- Visemes? FBFLayer?
    if moho:LayerAsSwitch(layer):InterpMode() then
        return 0
    end
    local group = moho:LayerAsGroup(layer)
    -- no masking test: Switch layers don't have masking inside
    local cnt = group:CountLayers()
    while cnt > 0 do
        cnt = cnt - 1
        local sub = group:Layer(cnt)
        if self:IsAnimated(moho, sub) then
            return 0
        end
        local tp = sub:LayerType()
        if tp ~= MOHO.LT_VECTOR and tp ~= MOHO.LT_IMAGE then
            return 0
        end
    end
    return 1
end

function FA_export_csv:Run(moho)

    local name = moho.document:Name()
    if name == nil then
        name = ""
    end
    local item = name:sub(-5):lower()

    if item == ".moho" or item == ".anme" then
        name = name:sub(1, -6)
    end

    if name == "" then
        name = "Untitled"
    end
 
    local layerNum = 0
    local layers = {}
    local pos = moho.document:CountLayers()

    while pos > 0 do
        pos = pos - 1
        local layer = moho.document:Layer(pos)
        local stepIn = false
        local tp = layer:LayerType()

        --steps into CSV top-level groups to keep the compatibility with the importer

        if tp == MOHO.LT_GROUP then
            local parent = moho:LayerAsGroup(layer)
            if parent:GetGroupMask() == MOHO.GROUP_MASK_NONE and not self:IsAnimated(moho, parent) then
                local cnt = parent:CountLayers()
                stepIn = true
                while cnt > 0 do
                    cnt = cnt - 1
                    if parent:Layer(cnt):LayerType() ~= MOHO.LT_SWITCH then
                        stepIn = false
                        break
                    end
                end
            end
        end

        if stepIn then
            local parent = moho:LayerAsGroup(layer)
            local cnt = parent:CountLayers()
            while cnt > 0 do
                cnt = cnt - 1
                local sub = parent:Layer(cnt)
                layers[layerNum] = {}
                layers[layerNum].layer = sub
                layers[layerNum].parent = parent
                layers[layerNum].visible = sub:IsVisible()
                layers[layerNum].compat = self:CheckSwitch(moho, sub)
                layers[layerNum].subs = nil
                layerNum = layerNum + 1
            end
        else
            layers[layerNum]= {}
            layers[layerNum].layer = layer
            layers[layerNum].parent = nil
            layers[layerNum].visible = layer:IsVisible(0)
            local compat = 0

            if tp == MOHO.LT_SWITCH then
                compat = self:CheckSwitch(moho, layer)
            elseif tp == MOHO.LT_VECTOR or tp == MOHO.LT_IMAGE then
                if not self:IsAnimated(moho, layer) then
                    compat = 2
                end
            end
            layers[layerNum].compat= compat
            layers[layerNum].subs = nil
            layerNum = layerNum + 1
        end
    end

    if layerNum == 0 then
        LM.GUI.Alert(LM.GUI.ALERT_WARNING, "No layers to save")
        return
    end

    FA_export_csv.filePath = LM.GUI.SaveFile("Select CSV file to save")
    if self.filePath == "" then
        return
    end

    if self.filePath:sub(-4):lower() ~= ".csv" then
        self.filePath = self.filePath..".csv"
    end

    local f = io.open(self.filePath, "w")

    if f == nil then
        LM.GUI.Alert(LM.GUI.ALERT_WARNING, "Sorry, cannot create file.", self.filePath)
        return
    end

    pos = self.filePath:len()
    local c

    while pos > 0 do
        c = self.filePath:sub(pos, pos)

        if c == "/" or c == "\\" then
            self.dirSep = c
            break
        end
        pos = pos - 1
    end
    local path = self.filePath:sub(1, -4).."frames"..c
    -- hoping this works on every systems: no other solution to create a directory
    os.execute("mkdir \""..path.."\"")

    local crlf = "\n"

    f:write("UTF-8, TVPaint, \"CSV 1.0\"", crlf)
    f:write("Project Name, Width, Height, Frame Count, Layer Count, Frame Rate, Pixel Aspect Ratio, Field Mode", crlf)

    f:write("\"", name, "\", ", moho.document:Width(), ", ", moho.document:Height(), ", ")
    f:write(moho.document:EndFrame() - moho.document:StartFrame() + 1, ", ", layerNum, ", ")
    f:write(string.format("%.6f", moho.document:Fps()), ", 1.000000, Progressive", crlf)

    layerNum = layerNum - 1

    f:write(crlf, "#Layers")
    for pos = 0, layerNum do
        f:write(", \"", layers[pos].layer:Name(), "\"")
    end
    f:write(crlf, "#Density")
    for pos = 0, layerNum do
        local alpha = layers[pos].layer.fAlpha:GetValue(0)
        if layers[pos].parent ~= nil then
            alpha = alpha * layers[pos].parent.fAlpha:GetValue(0)
        end
        f:write(", ", string.format("%.6f", alpha))
    end
    f:write(crlf, "#Blending")
    for pos = 0, layerNum do
        name = layers[pos].layer:BlendingMode()
        item = "Color"

        if name == MOHO.BM_MULTIPLY then
            item = "Multiply"
        elseif name == MOHO.BM_SCREEN then
            item = "Screen"
        elseif name == MOHO.BM_OVERLAY then
            item = "Overlay"
        elseif name == MOHO.BM_ADD then
            item = "Add"
        elseif name == MOHO.BM_DIFFERENCE then
            item = "Difference"
        elseif name == MOHO.BM_HUE then
            item = "Hue"
        elseif name == MOHO.BM_SATURATION then
            item = "Saturation"
        elseif name == MOHO.BM_LUMINOSITY then
            item = "Value"
        end
        f:write(", \"", item, "\"")
    end
    f:write(crlf, "#Visible")
    for pos = 0, layerNum do
        if layers[pos].visible then
            f:write(", 1")
        else
            f:write(", 0")
        end
    end

    local curFrame = moho.frame
    local frame = moho.document:StartFrame()

    for pos = 0, layerNum do
        layers[pos].layer:SetVisible(false)
    end

    while frame <= moho.document:EndFrame() do
        f:write(crlf, string.format("#%05d", frame - moho.document:StartFrame()))

        for pos = 0, layerNum do
            layer = layers[pos].layer
            name = string.format("frame%03d-%05d.png", pos, frame)

            if not layer.fVisibility:GetValue(frame) then
                name = ""
            elseif layers[pos].parent ~= nil then
                if not layers[pos].parent.fVisibility:GetValue(frame) then
                    name = ""
                end
            end

            if name ~= "" then
                --importer compatible blank frame test
                local vec = nil

                if layer:LayerType() == MOHO.LT_SWITCH then
                    --any switch layers, test for empty contents
                    local cnt = moho:LayerAsGroup(layer):CountLayers()
                    local selName = moho:LayerAsSwitch(layer):GetValue(frame)
                    while cnt > 0 do
                        cnt = cnt - 1
                        local sub = moho:LayerAsGroup(layer):Layer(cnt)
                        if sub:Name() == selName then
                            if not sub.fVisibility:GetValue(frame) then
                                name = ""
                            elseif sub:LayerType() == MOHO.LT_VECTOR then
                                vec = sub
                            end
                            break
                        end
                    end
                elseif layer:LayerType() == MOHO.LT_VECTOR then
                    vec = layer
                end

                if vec ~= nil then
                    if moho:LayerAsVector(vec):Mesh():CountPoints() == 0 then
                        -- blank frame
                        name = ""
                    end
                end
            end

            if name ~= "" then
                if layers[pos].compat == 1 then
                    --switch mode
                    if layers[pos].subs == nil then
                       layers[pos].subs = {}
                    end
                    local selName = moho:LayerAsSwitch(layer):GetValue(frame)
                    local sel = layers[pos].subs[selName]
                    if sel == nil then
                        layers[pos].subs[selName]= name
                        --render the frame
                        if moho.frame ~= frame then
                            moho:SetCurFrame(frame)
                            moho:UpdateUI()
                        end
                        layer:SetVisible(true)
                        moho:FileRender(path..name)
                        layer:SetVisible(false)
                    else
                        name = sel
                    end
                elseif layers[pos].compat == 2 then
                    --image mode
                    if layers[pos].subs == nil then
                        layers[pos].subs = name
                        --render the frame
                        if moho.frame ~= frame then
                            moho:SetCurFrame(frame)
                            moho:UpdateUI()
                        end
                        layer:SetVisible(true)
                        moho:FileRender(path..name)
                        layer:SetVisible(false)
                    else
                        name = layers[pos].subs
                    end
                else
                    --render the frame
                    if moho.frame ~= frame then
                        moho:SetCurFrame(frame)
                        moho:UpdateUI()
                    end
                    layer:SetVisible(true)
                    moho:FileRender(path..name)
                    layer:SetVisible(false)
                end
            end
            f:write(", \"",name,"\"")
        end
        frame = frame + 1
    end
    f:write(crlf)
    f:close()

    for pos = 0, layerNum do
        layers[pos].layer:SetVisible(layers[pos].visible)
    end
    moho:SetCurFrame(curFrame)
end
